@@ -1,6 +1,9 @@
 #include "tickit.h"
+#include "hooklists.h"
 
+#include <stdio.h>   /* sscanf */
 #include <stdlib.h>
+#include <string.h>
 
 struct TickitPen {
   signed   int fg      : 9, /* 0 - 255 or -1 */
@@ -24,14 +27,20 @@ struct TickitPen {
                  strike  : 1,
                  altfont : 1;
   } valid;
+
+  struct TickitEventHook *hooks;
 };
 
+DEFINE_HOOKLIST_FUNCS(pen,TickitPen,TickitPenEventFn)
+
 TickitPen *tickit_pen_new(void)
 {
   TickitPen *pen = malloc(sizeof(TickitPen));
   if(!pen)
     return NULL;
 
+  pen->hooks = NULL;
+
   for(TickitPenAttr attr = 0; attr < TICKIT_N_PEN_ATTRS; attr++)
     tickit_pen_clear_attr(pen, attr);
 
@@ -40,6 +49,7 @@ TickitPen *tickit_pen_new(void)
 
 void tickit_pen_destroy(TickitPen *pen)
 {
+  tickit_hooklist_unbind_and_destroy(pen->hooks, pen);
   free(pen);
 }
 
@@ -85,6 +95,10 @@ int tickit_pen_is_nondefault(TickitPen *pen)
       if(tickit_pen_get_int_attr(pen, attr) > -1)
         return 1;
       break;
+    case TICKIT_PENTYPE_COLOUR:
+      if(tickit_pen_get_colour_attr(pen, attr) > -1)
+        return 1;
+      break;
     }
   }
   return 0;
@@ -115,8 +129,9 @@ void tickit_pen_set_bool_attr(TickitPen *pen, TickitPenAttr attr, int val)
     case TICKIT_PEN_REVERSE: pen->reverse = !!val; pen->valid.reverse = 1; break;
     case TICKIT_PEN_STRIKE:  pen->strike  = !!val; pen->valid.strike  = 1; break;
     default:
-      break;
+      return;
   }
+  run_events(pen, TICKIT_EV_CHANGE, NULL);
 }
 
 int tickit_pen_get_int_attr(TickitPen *pen, TickitPenAttr attr)
@@ -125,8 +140,6 @@ int tickit_pen_get_int_attr(TickitPen *pen, TickitPenAttr attr)
     return -1;
 
   switch(attr) {
-    case TICKIT_PEN_FG:      return pen->fg;
-    case TICKIT_PEN_BG:      return pen->bg;
     case TICKIT_PEN_ALTFONT: return pen->altfont;
     default:
       return 0;
@@ -136,12 +149,75 @@ int tickit_pen_get_int_attr(TickitPen *pen, TickitPenAttr attr)
 void tickit_pen_set_int_attr(TickitPen *pen, TickitPenAttr attr, int val)
 {
   switch(attr) {
-    case TICKIT_PEN_FG:      pen->fg      = val; pen->valid.fg      = 1; break;
-    case TICKIT_PEN_BG:      pen->bg      = val; pen->valid.bg      = 1; break;
     case TICKIT_PEN_ALTFONT: pen->altfont = val; pen->valid.altfont = 1; break;
     default:
-      break;
+      return;
   }
+  run_events(pen, TICKIT_EV_CHANGE, NULL);
+}
+
+/* Cheat and pretend the index of a colour attribute is a number attribute */
+int tickit_pen_get_colour_attr(TickitPen *pen, TickitPenAttr attr)
+{
+  if(!tickit_pen_has_attr(pen, attr))
+    return -1;
+
+  switch(attr) {
+    case TICKIT_PEN_FG: return pen->fg;
+    case TICKIT_PEN_BG: return pen->bg;
+    default:
+      return 0;
+  }
+}
+
+void tickit_pen_set_colour_attr(TickitPen *pen, TickitPenAttr attr, int val)
+{
+  switch(attr) {
+    case TICKIT_PEN_FG: pen->fg = val; pen->valid.fg = 1; break;
+    case TICKIT_PEN_BG: pen->bg = val; pen->valid.bg = 1; break;
+    default:
+      return;
+  }
+  run_events(pen, TICKIT_EV_CHANGE, NULL);
+}
+
+static const char *colournames[] = {
+  "black",
+  "red",
+  "green",
+  "yellow",
+  "blue",
+  "magenta",
+  "cyan",
+  "white",
+};
+
+int tickit_pen_set_colour_attr_desc(TickitPen *pen, TickitPenAttr attr, const char *desc)
+{
+  int hi = 0;
+  int val;
+  if(strncmp(desc, "hi-", 3) == 0) {
+    desc += 3;
+    hi   = 8;
+  }
+
+  if(sscanf(desc, "%d", &val) == 1) {
+    if(hi && val > 7)
+      return 0;
+
+    tickit_pen_set_colour_attr(pen, attr, val + hi);
+    return 1;
+  }
+
+  for(val = 0; val < sizeof(colournames)/sizeof(colournames[0]); val++) {
+    if(strcmp(desc, colournames[val]) != 0)
+      continue;
+
+    tickit_pen_set_colour_attr(pen, attr, val + hi);
+    return 1;
+  }
+
+  return 0;
 }
 
 void tickit_pen_clear_attr(TickitPen *pen, TickitPenAttr attr)
@@ -157,8 +233,9 @@ void tickit_pen_clear_attr(TickitPen *pen, TickitPenAttr attr)
     case TICKIT_PEN_ALTFONT: pen->valid.altfont = 0; break;
 
     case TICKIT_N_PEN_ATTRS:
-      break;
+      return;
   }
+  run_events(pen, TICKIT_EV_CHANGE, NULL);
 }
 
 int tickit_pen_equiv_attr(TickitPen *a, TickitPen *b, TickitPenAttr attr)
@@ -168,6 +245,8 @@ int tickit_pen_equiv_attr(TickitPen *a, TickitPen *b, TickitPenAttr attr)
     return tickit_pen_get_bool_attr(a, attr) == tickit_pen_get_bool_attr(b, attr);
   case TICKIT_PENTYPE_INT:
     return tickit_pen_get_int_attr(a, attr) == tickit_pen_get_int_attr(b, attr);
+  case TICKIT_PENTYPE_COLOUR:
+    return tickit_pen_get_colour_attr(a, attr) == tickit_pen_get_colour_attr(b, attr);
   }
 
   return 0;
@@ -182,16 +261,77 @@ void tickit_pen_copy_attr(TickitPen *dst, TickitPen *src, TickitPenAttr attr)
   case TICKIT_PENTYPE_INT:
     tickit_pen_set_int_attr(dst, attr, tickit_pen_get_int_attr(src, attr));
     return;
+  case TICKIT_PENTYPE_COLOUR:
+    tickit_pen_set_colour_attr(dst, attr, tickit_pen_get_colour_attr(src, attr));
+    return;
   }
 
   return;
 }
 
+void tickit_pen_copy(TickitPen *dst, TickitPen *src, int overwrite)
+{
+  int changed = 0;
+  for(TickitPenAttr attr = 0; attr < TICKIT_N_PEN_ATTRS; attr++) {
+    if(!tickit_pen_has_attr(src, attr))
+      continue;
+    if(tickit_pen_has_attr(dst, attr) && !overwrite)
+      continue;
+    if(tickit_pen_equiv_attr(src, dst, attr))
+      continue;
+
+    /* Avoid using copy_attr so it doesn't invoke change events yet */
+    switch(attr) {
+    case TICKIT_PEN_FG:
+      dst->fg = src->fg;
+      dst->valid.fg = 1;
+      break;
+    case TICKIT_PEN_BG:
+      dst->bg = src->bg;
+      dst->valid.bg = 1;
+      break;
+    case TICKIT_PEN_BOLD:
+      dst->bold = src->bold;
+      dst->valid.bold = 1;
+      break;
+    case TICKIT_PEN_ITALIC:
+      dst->italic = src->italic;
+      dst->valid.italic = 1;
+      break;
+    case TICKIT_PEN_UNDER:
+      dst->under = src->under;
+      dst->valid.under = 1;
+      break;
+    case TICKIT_PEN_REVERSE:
+      dst->reverse = src->reverse;
+      dst->valid.reverse = 1;
+      break;
+    case TICKIT_PEN_STRIKE:
+      dst->strike = src->strike;
+      dst->valid.strike = 1;
+      break;
+    case TICKIT_PEN_ALTFONT:
+      dst->altfont = src->altfont;
+      dst->valid.altfont = 1;
+      break;
+    case TICKIT_N_PEN_ATTRS:
+      continue;
+    }
+
+    changed++;
+  }
+
+  if(changed)
+    run_events(dst, TICKIT_EV_CHANGE, NULL);
+}
+
 TickitPenAttrType tickit_pen_attrtype(TickitPenAttr attr)
 {
   switch(attr) {
     case TICKIT_PEN_FG:
     case TICKIT_PEN_BG:
+      return TICKIT_PENTYPE_COLOUR;
+
     case TICKIT_PEN_ALTFONT:
       return TICKIT_PENTYPE_INT;
 
